from typing import Callable

import numpy as np
from numba import jit
from numpy import log

from pyfracgen.result import Result


@jit  # type: ignore[misc]
def _julia(
    c: float,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    maxiter: int = 100,
    horizon: float = 2.0**40,
    log_smooth: bool = True,
) -> tuple[np.ndarray, int, int, int]:

    xmin, xmax = [float(xbound[0]), float(xbound[1])]
    ymin, ymax = [float(ybound[0]), float(ybound[1])]

    nx = width * dpi
    ny = height * dpi

    xvals = np.array(
        [xmin + i * (xmax - xmin) / nx for i in range(nx)], dtype=np.float64
    )
    yvals = np.array(
        [ymin + i * (ymax - ymin) / ny for i in range(ny)], dtype=np.float64
    )

    lattice = np.zeros((int(nx), int(ny)), dtype=np.float64)

    log_horizon = log(log(horizon)) / log(2)

    for i in range(len(xvals)):
        for j in range(len(yvals)):

            z = xvals[i] + 1j * yvals[j]

            for iteration in range(maxiter):

                az = abs(z)

                if az > horizon:
                    if log_smooth:
                        lattice[i, j] = iteration - log(log(az)) / log(2) + log_horizon
                    else:
                        lattice[i, j] = iteration
                    break

                z = update_func(z, c)

    return (lattice.T, width, height, dpi)


def julia_series(
    c_vals: np.ndarray,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    maxiter: int = 100,
    horizon: float = 2.0**40,
    log_smooth: bool = True,
) -> list[Result]:

    series = []

    for c in c_vals:
        res = _julia(
            c,
            xbound,
            ybound,
            update_func,
            width=width,
            height=height,
            dpi=dpi,
            maxiter=maxiter,
            horizon=horizon,
            log_smooth=log_smooth,
        )
        res = Result(*res)
        series.append(res)

    return series


def julia(
    c: float,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    maxiter: int = 100,
    horizon: float = 2.0**40,
    log_smooth: bool = True,
) -> Result:

    res = _julia(
        c,
        xbound,
        ybound,
        update_func,
        width=width,
        height=height,
        dpi=dpi,
        maxiter=maxiter,
        horizon=horizon,
        log_smooth=log_smooth,
    )

    return Result(*res)



from .buddhabrot import buddhabrot, compute_cvals
from .images import images
from .julia import julia, julia_series
from .lyapunov import lyapunov
from .mandelbrot import mandelbrot
from .randomwalk import construct_moves, random_walk
from .result import Result
from .updaters import funcs

__all__ = [
    "buddhabrot",
    "compute_cvals",
    "images",
    "julia",
    "julia_series",
    "lyapunov",
    "mandelbrot",
    "random_walk",
    "construct_moves",
    "Result",
    "funcs",
]
__version__ = "0.0.9"



from itertools import combinations

import numpy as np
from numba import jit
from numpy import array
from numpy.random import randint

from pyfracgen.result import Result


def construct_moves(basis: np.ndarray) -> np.ndarray:

    basis = np.r_[basis, -1 * basis, [array([0, 0, 0])]]
    moves = np.unique(array([b0 + b1 for b0, b1 in combinations(basis, 2)]), axis=0)
    moves = array([m for m in moves if np.any(m)])

    return moves


@jit  # type: ignore[misc]
def _random_walk(
    moves: np.ndarray,
    niter: int,
    width: int = 5,
    height: int = 5,
    depth: int = 1,
    dpi: int = 100,
    tracking: str = "visitation",
) -> tuple[np.ndarray, int, int, int]:

    lattice = np.zeros(
        (int(height * dpi), int(width * dpi), int(depth)), dtype=np.float32
    )
    shape = array([height * dpi, width * dpi, depth])
    nmoves = len(moves)

    l0, l1, l2 = shape
    indices = array([height * dpi, width * dpi, depth]) / 2.0

    for iteration in range(niter):

        move = moves[randint(0, nmoves)]
        indices += move
        i, j, k = int(indices[0] % l0), int(indices[1] % l1), int(indices[2] % l2)

        if tracking == "visitation":
            lattice[i, j, k] += 1.0
        elif tracking == "temporal":
            if lattice[i, j, k] == 0.0:
                lattice[i, j, k] += iteration

    lattice /= np.amax(lattice)

    return (lattice, width, height, dpi)


def random_walk(
    moves: np.ndarray,
    niter: int,
    width: int = 5,
    height: int = 5,
    depth: int = 1,
    dpi: int = 100,
    tracking: str = "visitation",
) -> Result:

    res = _random_walk(
        moves,
        niter,
        width=width,
        height=height,
        depth=depth,
        dpi=dpi,
        tracking=tracking,
    )

    return Result(*res)



from cmath import cos, exp, sin

from numba import jit
from numpy import conj


@jit  # type: ignore[misc]
def power(z: complex, c: complex, n: int = 2) -> complex:
    return z**n + c


@jit  # type: ignore[misc]
def conj_power(z: complex, c: complex, n: int = 2) -> complex:
    val: complex = conj(z) ** n + c
    return val


@jit  # type: ignore[misc]
def cosine(z: complex, c: complex) -> complex:
    return c * cos(z)


@jit  # type: ignore[misc]
def sine(z: complex, c: float) -> complex:
    return c * sin(z)


@jit  # type: ignore[misc]
def exponential(z: complex, c: complex) -> complex:
    return c * exp(z)


@jit  # type: ignore[misc]
def magnetic_1(z: complex, c: complex) -> complex:

    t0 = (z * z + c - 1) / (2 * z + c - 2)
    t1 = (z * z + c - 1) / (2 * z + c - 2)

    return t0 * t1


@jit  # type: ignore[misc]
def magnetic_2(z: complex, c: complex) -> complex:

    t0 = (z * z * z * 3 * (c - 1) * z + (c - 1) * (c - 2)) / (
        3 * z * z + 3 * (c - 2) * z + (c - 1) * (c - 2) + 1
    )
    t1 = (z * z * z * 3 * (c - 1) * z + (c - 1) * (c - 2)) / (
        3 * z * z + 3 * (c - 2) * z + (c - 1) * (c - 2) + 1
    )

    return t0 * t1



from typing import Callable

import numpy as np
from numba import jit
from numpy.random import random

from pyfracgen.mandelbrot import _mandelbrot
from pyfracgen.result import Result


@jit  # type: ignore[misc]
def compute_cvals(
    n_cvals: int,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    importance_weight: float = 0.75,
    transpose_energy_grid: bool = False,
) -> np.ndarray:

    xmin, xmax = [float(xbound[0]), float(xbound[1])]
    ymin, ymax = [float(ybound[0]), float(ybound[1])]

    nx = width * dpi
    ny = height * dpi

    xvals = np.array(
        [xmin + i * (xmax - xmin) / nx for i in range(nx)], dtype=np.float32
    )
    yvals = np.array(
        [ymin + i * (ymax - ymin) / ny for i in range(ny)], dtype=np.float32
    )

    xboxs = [(xvals[i], xvals[i + 1]) for i in range(len(xvals) - 1)]
    yboxs = [(yvals[i], yvals[i + 1]) for i in range(len(yvals) - 1)]
    xboxs = xboxs + [(xboxs[-1][1], xmax)]
    yboxs = yboxs + [(yboxs[-1][1], ymax)]

    nr = int(round(n_cvals * (1 - importance_weight)))
    cvals = []

    # randomly sampled starting points
    for k in range(nr):

        c = xmin + (random() * (xmax - xmin)) + 1j * (ymin + (random() * (ymax - ymin)))
        cvals.append(c)

    # energy grid sampled starting points
    if importance_weight > 0.0:

        ni = int(round(n_cvals * importance_weight))

        energy_grid = _mandelbrot(
            xbound,
            ybound,
            update_func,
            width=width,
            height=height,
            dpi=dpi,
            maxiter=1000,
            horizon=2.5,
            log_smooth=False,
        )[0].T

        energy_grid = (energy_grid / energy_grid.sum()) * ni
        energy_grid = energy_grid.T if transpose_energy_grid else energy_grid

        for i in range(nx):
            for j in range(ny):

                num = int(round(energy_grid[i, j]))
                xlo, xhi = xboxs[i]
                ylo, yhi = yboxs[j]
                cs = (
                    xlo
                    + (random(num) * (xhi - xlo))
                    + 1j * (ylo + (random(num) * (yhi - ylo)))
                )
                cvals.extend(list(cs))

    return np.array(cvals)


@jit  # type: ignore[misc]
def _buddhabrot(
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    cvals: np.ndarray,
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    maxiter: int = 100,
    horizon: float = 1.0e6,
) -> tuple[np.ndarray, int, int, int]:

    xmin, xmax = [float(xbound[0]), float(xbound[1])]
    ymin, ymax = [float(ybound[0]), float(ybound[1])]

    nx = width * dpi
    ny = height * dpi

    xvals = np.array(
        [xmin + i * (xmax - xmin) / nx for i in range(nx)], dtype=np.float32
    )
    yvals = np.array(
        [ymin + i * (ymax - ymin) / ny for i in range(ny)], dtype=np.float32
    )
    xboxs = [(xvals[i], xvals[i + 1]) for i in range(len(xvals) - 1)]
    yboxs = [(yvals[i], yvals[i + 1]) for i in range(len(yvals) - 1)]
    xboxs = xboxs + [(xboxs[-1][1], xmax)]
    yboxs = yboxs + [(yboxs[-1][1], ymax)]

    lattice = np.zeros((int(width * dpi), int(height * dpi)), dtype=np.float32)

    for c in cvals:

        z = c
        trial_sequence = []
        sequence = []

        for iteration in range(maxiter):

            az = np.abs(z)
            trial_sequence.append(z)

            if az > horizon:
                sequence.extend(trial_sequence)
                break

            z = update_func(z, c)

        for c in sequence:

            indx = 0
            indy = 0

            for bx in range(nx):
                if xboxs[bx][0] < c.real < xboxs[bx][1]:
                    indx += bx
                    break

            for by in range(ny):
                if yboxs[by][0] < c.imag < yboxs[by][1]:
                    indy += by
                    break

            if indx != 0 and indy != 0:
                lattice[indx, indy] += 1

    return (lattice.T, width, height, dpi)


def buddhabrot(
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    cvals: np.ndarray,
    update_func: Callable[[complex, complex], complex],
    width: int = 5,
    height: int = 5,
    dpi: int = 100,
    maxiter: int = 100,
    horizon: float = 1.0e6,
) -> Result:

    res = _buddhabrot(
        xbound,
        ybound,
        cvals,
        update_func,
        width=width,
        height=height,
        dpi=dpi,
        maxiter=maxiter,
        horizon=horizon,
    )

    return Result(*res)



from __future__ import annotations

import pickle
from dataclasses import dataclass
from pathlib import Path

import numpy as np


@dataclass
class Result:

    image_array: np.ndarray
    width_inches: float
    height_inches: float
    dpi: int

    @classmethod
    def load(cls, file: Path) -> Result:
        with open(file, "rb") as f:
            res = pickle.load(f)
        return cls(*res)

    def save(self, name: Path = Path("save")) -> None:
        res = [self.image_array, self.width_inches, self.height_inches, self.dpi]
        with open(name, "wb") as f:
            pickle.dump(res, f)



import numpy as np
from numba import jit

from pyfracgen.result import Result


@jit  # type: ignore[misc]
def _lyapunov(
    string: str,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    n_init: int = 200,
    n_iter: int = 800,
    width: int = 3,
    height: int = 3,
    dpi: int = 100,
) -> tuple[np.ndarray, int, int, int]:

    instructions = [0 if s == "A" else 1 for s in string]
    xmin, xmax = [float(xbound[0]), float(xbound[1])]
    ymin, ymax = [float(ybound[0]), float(ybound[1])]

    nx = width * dpi
    ny = height * dpi

    xvals = np.array(
        [xmin + i * (xmax - xmin) / nx for i in range(nx)], dtype=np.float64
    )
    yvals = np.array(
        [ymin + i * (ymax - ymin) / ny for i in range(ny)], dtype=np.float64
    )

    length = len(instructions)
    lattice = np.zeros((int(nx), int(ny)), dtype=np.float64)

    for i in range(len(xvals)):
        for j in range(len(yvals)):

            coord = [xvals[i], yvals[j]]
            n = 0

            x = 0.5
            for _init in range(n_init):

                rn = coord[instructions[n % length]]
                x = (rn * x) * (1 - x)
                n += 1

            lamd = 0.0
            for _iter in range(n_iter):

                rn = coord[instructions[n % length]]
                x = (rn * x) * (1 - x)
                lamd += np.log(np.abs(rn - 2 * rn * x))
                n += 1

            lattice[i, j] += lamd / n_iter

    return (lattice.T, width, height, dpi)


def lyapunov(
    string: str,
    xbound: tuple[float, float],
    ybound: tuple[float, float],
    n_init: int = 200,
    n_iter: int = 800,
    width: int = 3,
    height: int = 3,
    dpi: int = 100,
) -> Result:

    res = _lyapunov(
        string,
        xbound,
        ybound,
        n_init=n_init,
        n_iter=n_iter,
        width=width,
        height=height,
        dpi=dpi,
    )

    return Result(*res)



import matplotlib.animation as animation
import numpy as np
from matplotlib import colors
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from numpy.ma import masked_where

from pyfracgen.result import Result


def stack_cmaps(cmap: colors.Colormap, n_stacks: int) -> colors.LinearSegmentedColormap:

    colorline = np.array(cmap(np.linspace(0, 1, 200)))

    for _ in range(n_stacks - 1):
        colorline = np.vstack((colorline, cmap(np.linspace(0, 1, 200))))

    mymap = colors.LinearSegmentedColormap.from_list("my_colormap", colorline)

    return mymap


def image(
    res: Result,
    cmap: colors.Colormap = plt.cm.hot,
    ticks: str = "off",
    gamma: float = 0.3,
    vert_exag: float = 0,
    ls: tuple[int, int] = (315, 10),
) -> tuple[Figure, plt.Axes]:

    arr = res.image_array
    width = res.width_inches
    height = res.height_inches
    dpi = res.dpi

    w, h = plt.figaspect(arr)
    fig, ax0 = plt.subplots(figsize=(w, h), dpi=dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    plt.axis(ticks)

    norm = colors.PowerNorm(gamma)
    light = colors.LightSource(azdeg=ls[0], altdeg=ls[1])

    if vert_exag != 0.0:
        ls = light.shade(
            arr, cmap=cmap, norm=norm, vert_exag=vert_exag, blend_mode="hsv"
        )
        ax0.imshow(ls, origin="lower")
    else:
        ax0.imshow(arr, origin="lower", cmap=cmap, norm=norm)

    fs = plt.gcf()
    fs.set_size_inches(width, height)

    return fig, ax0


def nebula_image(
    res_blue: Result,
    res_green: Result,
    res_red: Result,
    ticks: str = "off",
    gamma: float = 1.0,
) -> tuple[Figure, plt.Axes]:

    arr_blue = res_blue.image_array
    width = res_blue.width_inches
    height = res_blue.height_inches
    dpi = res_blue.dpi
    arr_green = res_green.image_array
    arr_red = res_red.image_array

    arr_blue /= np.amax(arr_blue)
    arr_green /= np.amax(arr_green)
    arr_red /= np.amax(arr_red)

    final = np.dstack((arr_red, arr_green, arr_blue))

    w, h = plt.figaspect(arr_blue)
    fig, ax0 = plt.subplots(figsize=(w, h), dpi=dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    plt.axis(ticks)
    fs = plt.gcf()
    fs.set_size_inches(width, height)
    ax0.imshow(final**gamma, origin="lower")

    return fig, ax0


def markus_lyapunov_image(
    res: Result,
    cmap_negative: colors.Colormap,
    cmap_positive: colors.Colormap,
    gammas: tuple[float, float] = (1.0, 1.0),
    ticks: str = "off",
) -> tuple[Figure, plt.Axes]:

    arr = res.image_array
    width = res.width_inches
    height = res.height_inches
    dpi = res.dpi

    neg = masked_where(arr > 0.0, arr)
    pos = masked_where(arr < 0.0, arr)

    w, h = plt.figaspect(neg)
    fig, ax0 = plt.subplots(figsize=(width, height), dpi=dpi)
    ax0.imshow(
        neg, cmap=cmap_negative, origin="lower", norm=colors.PowerNorm(gammas[0])
    )
    ax0.imshow(
        pos, cmap=cmap_positive, origin="lower", norm=colors.PowerNorm(gammas[1])
    )

    fig.subplots_adjust(0, 0, 1, 1)
    plt.axis(ticks)
    fs = plt.gcf()
    fs.set_size_inches(width, height)

    return fig, ax0


def random_walk_image(
    res: Result,
    cmap: colors.Colormap = plt.cm.hot,
    ticks: str = "off",
    gamma: float = 0.3,
    alpha_scale: float = 1.0,
) -> tuple[Figure, plt.Axes]:

    arr = res.image_array
    width = res.width_inches
    height = res.height_inches
    dpi = res.dpi

    w, h = plt.figaspect(arr[:, :, 0])
    fig, ax0 = plt.subplots(figsize=(w, h), dpi=dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    plt.axis(ticks)
    max_ind = float(arr.shape[-1] + 1)

    for i in range(arr.shape[-1]):

        im = arr[..., i]
        im = masked_where(im == 0, im)
        alpha = 1 - (i + 1) / max_ind
        alpha *= alpha_scale
        norm = colors.PowerNorm(gamma)
        ax0.imshow(
            im, origin="lower", alpha=alpha, cmap=cmap, norm=norm, interpolation=None
        )

    fs = plt.gcf()
    fs.set_size_inches(width, height)

    return fig, ax0


def save_animation(
    series: list[Result],
    fps: int = 15,
    bitrate: int = 1800,
    cmap: int = plt.cm.hot,
    filename: str = "ani",
    ticks: str = "off",
    gamma: float = 0.3,
    vert_exag: float = 0,
    ls: tuple[int, int] = (315, 10),
) -> None:

    width = series[0].width_inches
    height = series[0].height_inches
    dpi = series[0].dpi

    fig = plt.figure()
    fig.subplots_adjust(0, 0, 1, 1)
    fs = plt.gcf()
    fs.set_size_inches(width, height)
    plt.axis(ticks)

    writer = animation.PillowWriter(
        fps=fps, metadata=dict(artist="Me"), bitrate=bitrate
    )
    norm = colors.PowerNorm(gamma)
    light = colors.LightSource(azdeg=ls[0], altdeg=ls[1])

    ims = []
    for s in series:

        arr = s.image_array
        ls = light.shade(
            arr, cmap=cmap, norm=norm, vert_exag=vert_exag, blend_mode="hsv"
        )
        im = plt.imshow(ls, origin="lower", norm=norm)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True, repeat_delay=1000)
    ani.save(f"{filename}.gif", dpi=dpi, writer=writer)
